package net.pingex.discordbot;

import org.apache.commons.lang3.ClassUtils;
import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import sx.blah.discord.api.EventSubscriber;
import sx.blah.discord.handle.impl.events.MessageReceivedEvent;
import sx.blah.discord.util.DiscordException;
import sx.blah.discord.util.HTTP429Exception;
import sx.blah.discord.util.MissingPermissionsException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.*;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class parses and dispatch Discord commands
 * @version 0.1-dev
 * @author Raphael "Pingex" NAAS
 */
class CommandDispatcher
{
	private Logger logger;

	/**
	 * Singleton instance
	 */
	private static final CommandDispatcher INSTANCE = new CommandDispatcher();

	/**
	 * Contains all available commands, built using `rebuildCommandList()`
	 */
	private Map<String, InvokableMethod> commandList;

	/**
	 * Contains all the shorthanded commands
	 */
	private Map<String, String> shortList;

	/**
	 * Thread pool used to invoke commands
	 */
	private ExecutorService threadPool;

	/**
	 * Basic Constructor, automatically rebuilds command list from the modules registry
	 */
	private CommandDispatcher()
	{
		logger = Logger.getLogger(this.getClass().getName());
		rebuildCommandList();
		threadPool = Executors.newCachedThreadPool(new BasicThreadFactory.Builder().namingPattern("CommandDispatcher-%d").build()); // TODO: threadPool.shutdown()

		if(!Configuration.exists("general", "commandPrefix") || Configuration.getValue("general", "commandPrefix").isEmpty())
		{
			logger.warning("Invalid command prefix detected, falling back to \"!\"");
			Configuration.setValue("general", "commandPrefix", "!");
		}
	}

	/**
	 * Gets the unique instance of `CommandDispatcher`
	 * @return The current instance of `CommandDispatcher`
	 */
	public static CommandDispatcher getInstance()
	{
		return INSTANCE;
	}

	/**
	 * Fired when receiving a message
	 * @param event Event from Discord API
	 */
	@EventSubscriber
	public void onMessageReceivedEvent(MessageReceivedEvent event)
	{
		// Command matcher
		Matcher m = Pattern.compile("^" + Configuration.getValue("general", "commandPrefix") + "([:\\w]+)(?: (.*))?$").matcher(event.getMessage().getContent());
		if(!m.matches()) return; // We don't need to go further if it's not even a command
		String fullCommand = m.group(1).toLowerCase();

		// Arg splitter
		Matcher am = Pattern.compile("([^\"]\\S*|\".+?\")\\s*").matcher((m.group(2) != null) ? m.group(2) : "");
		ArrayList<String> args = new ArrayList<>();
		while(am.find()) args.add(am.group(1).replace("\"", ""));

		// Shorthand matcher
		if(!fullCommand.contains(":") && shortList.containsKey(fullCommand))
			fullCommand = shortList.get(fullCommand);

		String commandAnswer = null;
		Object[] parsedArray = null;

		// Conditions
		if(commandList.containsKey(fullCommand))
		{
			logger.info("Command invoked (" + event.getMessage().getAuthor().getName() + "#" + event.getMessage().getAuthor().getDiscriminator() + "): " + fullCommand);

			InvokableMethod foundMethod = commandList.get(fullCommand);
			parsedArray = new Object[foundMethod.getMethod().getParameterCount()];
			parsedArray[0] = event;

			// To be redone, maybe ?
			Boolean canOverrideRun = PermissionsModule.getInstance().canRun(event.getMessage().getGuild(), event.getMessage().getAuthor(), fullCommand);
			if(!foundMethod.getMethod().getAnnotation(Command.class).permission().eval(event))
				commandAnswer = "Permission denied.";
			if(canOverrideRun != null)
				commandAnswer = canOverrideRun ? null : "Permission denied.";

			if(commandAnswer == null)
			{
				if(foundMethod.getMethod().getParameterCount()-1 >= args.size())
				{
					for(int i=1; i < foundMethod.getMethod().getParameterCount(); i++)
					{
						try
						{
							parsedArray[i] = parse(foundMethod.getMethod().getParameterTypes()[i], args.size() > i-1 ? args.get(i-1) : null);
						} catch (IllegalArgumentException e)
						{
							commandAnswer = "Invalid arguments. Call `" + Configuration.getValue("general", "commandPrefix") + "help " + fullCommand + "` for help.";
						}

						// Try to know if the field is required (or not) from the Command annotation
						boolean isRequired = true;
						try { isRequired = foundMethod.getMethod().getAnnotation(Command.class).required()[i-1]; } catch (ArrayIndexOutOfBoundsException e) {}

						// if(ArgumentIsNull && (IsRequired || ArgumentIsRealPrimitiveOrObject))
						if(parsedArray[i] == null && (isRequired ||
								!(ClassUtils.isPrimitiveWrapper(foundMethod.getMethod().getParameterTypes()[i]) || foundMethod.getMethod().getParameterTypes()[i] == String.class)))
						{
							commandAnswer = "Invalid arguments. Call `" + Configuration.getValue("general", "commandPrefix") + "help " + fullCommand + "` for help.";
							break;
						}
					}
				}
				else
					commandAnswer = "Invalid arguments. Call `" + Configuration.getValue("general", "commandPrefix") + "help " + fullCommand + "` for help.";
			}
		}
		else
			commandAnswer = "Unknown command. Call `" + Configuration.getValue("general", "commandPrefix") + "list` to see all available commands.";

		// Run command
		if(commandAnswer == null)
		{
			String finalFullCommand = fullCommand;
			Object[] finalParsedArray = parsedArray;
			threadPool.submit(() -> {
				try
				{
					String ans = (String) commandList.get(finalFullCommand).invoke(finalParsedArray);
					if(ans != null) event.getMessage().getChannel().sendMessage("```\n" + ans + "\n```");
				} catch (IllegalArgumentException | IllegalAccessException e)
				{
					logger.severe("Couldn't call target method (" + e.getClass().getName() + "): " + e.getMessage());
				} catch (InvocationTargetException e)
				{
					logger.severe("An error occurred inside the target command: " + e.getCause().getMessage());
					e.getCause().printStackTrace();
				} catch (MissingPermissionsException | HTTP429Exception | DiscordException e)
				{
					logger.warning("Couldn't reply to command (" + e.getClass().getName() + "): " + e.getMessage());
				}
			});
		}
		else // Error answer
			try
			{
				event.getMessage().getChannel().sendMessage("```\n" + commandAnswer + "\n```");
			} catch (MissingPermissionsException | HTTP429Exception | DiscordException e)
			{
				logger.warning("Couldn't reply to command (" + e.getClass().getName() + "): " + e.getMessage());
			}
	}

	/**
	 * Rebuilds command list using the module registry.
	 * - It looks for methods with @Command annotation inside class with @Controllable annotation
	 * - Check his prototype (should be `method(IDiscordClient, ...):String`)
	 * - Add usable methods in his local registry, using InvokableMethod
	 * @see InvokableMethod
	 * @see ModulesRegistry
	 */
	public void rebuildCommandList()
	{
		logger.info("Rebuilding command list...");

		ArrayList<AbstractModule> registry = ModulesRegistry.getRegistry();
		commandList = new TreeMap<>();
		shortList = new TreeMap<>();

		int count = 0;
		int invalidCount = 0;

		for(AbstractModule i : registry)
			if(i.getClass().isAnnotationPresent(Controllable.class))
				for(Method j : i.getClass().getDeclaredMethods())
					if(j.isAnnotationPresent(Command.class))
					{
						String id = i.getClass().getAnnotation(Controllable.class).name() + ":" + j.getName().toLowerCase();

						if(j.getParameterCount() >= 1 && j.getParameterTypes()[0] == MessageReceivedEvent.class && j.getReturnType() == String.class)
							if(!commandList.containsKey(id))
							{
								commandList.put(id, new InvokableMethod(j, i));
								count++;

								// Eventual command shorthands
								if(!j.getAnnotation(Command.class).shorthand().isEmpty())
									if(!shortList.containsKey(j.getAnnotation(Command.class).shorthand().toLowerCase()))
										shortList.put(j.getAnnotation(Command.class).shorthand().toLowerCase(), id);
									else
										logger.warning("Conflicting shorthand for command " + id);
							}
							else
							{
								logger.warning("Conflicting command " + id);
								invalidCount++;
							}
						else
						{
							logger.warning("Command [" + id + "]: incorrect function prototype, thus won't be added to the command list.");
							invalidCount++;
						}
					}

		logger.info("Found " + count + " valid commands and " + invalidCount + " invalid commands.");
		logger.info("... Done");
	}

	/**
	 * Parses a String into a primitive variable. Choice between boolean, byte, short, int, long, float, double, string
	 * @param target Target primitive
	 * @param value Input value
	 * @return `null` if the value is null or the string content is "null", or the parsed value.
	 * @throws IllegalArgumentException If the target primitive/class isn't part of the previous list
	 */
	public static Object parse(Class<?> target, String value) throws IllegalArgumentException
	{
		if(value         == null   || value.equals("null")  ) return null;
		if(Boolean.class == target || Boolean.TYPE == target) return Boolean.parseBoolean(value);
		if(Byte.class    == target || Byte.TYPE    == target) return Byte.parseByte(value);
		if(Short.class   == target || Short.TYPE   == target) return Short.parseShort(value);
		if(Integer.class == target || Integer.TYPE == target) return Integer.parseInt(value);
		if(Long.class    == target || Long.TYPE    == target) return Long.parseLong(value);
		if(Float.class   == target || Float.TYPE   == target) return Float.parseFloat(value);
		if(Double.class  == target || Double.TYPE  == target) return Double.parseDouble(value);
		if(String.class  == target)                           return value;
		throw new IllegalArgumentException("Unknown target type");
	}

	/**
	 * Gets the command list
	 * @return The current command list
	 */
	public Map<String, InvokableMethod> getCommandList()
	{
		return commandList;
	}

	/**
	 * Gets the shorthands list
	 * @return The current short list
	 */
	public Map<String, String> getShortList()
	{
		return shortList;
	}
}
